@file:OptIn(ExperimentalUnsignedTypes::class)

package com.icodici.ubdata

import net.sergeych.bintools.ArrayDataSink
import net.sergeych.bintools.DataSink
import net.sergeych.bintools.DataSource
import net.sergeych.bintools.toDataSource

/**
 * Variable-length long integer encoding. the MSB (0x80) bit of each byte flags
 * that it is not the last one, and all ncecssary bits are encoded with 7-bit
 * portions, LSB to MSB (big endian of sorts).
 *
 * There is slower but more compact encoding variant, [Smartint] that is better when
 * encoding numbers that needs more than 22 bits. With smaller numbers its either
 * same or even worse, see [Smartint] docs.
 */
object Varint {
    fun encodeSigned(value: Long, sink: DataSink): Unit {
        var sigBit: ULong
        var x: ULong
        if (value < 0) {
            x = (-value).toULong()
            sigBit = 1u
        } else {
            x = value.toULong()
            sigBit = 0u
        }
        encode((x shl 1) or sigBit, sink)
    }

    fun encode(value: ULong, dout: DataSink) {
        var rest = value
        do {
            val x = (rest and 127u).toInt()
            rest = rest shr 7
            if (rest > 0u)
                dout.writeByte(x or 0x80)
            else
                dout.writeByte(x)

        } while (rest > 0u)
    }

    fun decode(source: DataSource): ULong {
        var result: ULong = 0u
        var count = 0
        while (true) {
            val x = source.readUByte().toInt()
            result = result or ((x and 0x7F).toULong() shl count)
            if ((x and 0x80) == 0)
                break
            count += 7
        }
        return result
    }


    fun decodeSigned(source: DataSource): Long {
        val x = decode(source)
        val result = (x shr 1).toLong()
        return if ((x and 1u).toInt() != 0) -result else result
    }

    fun encode(value: ULong): ByteArray {
        return ArrayDataSink().also { encode(value, it) }.toByteArray()
    }

    fun decode(packed: ByteArray) = decode(packed.toDataSource())

    fun encodeSigned(value: Long): ByteArray {
        return ArrayDataSink().also { encodeSigned(value, it) }.toByteArray()
    }

    fun decodeSigned(data: ByteArray): Long {
        return decodeSigned(data.toDataSource())
    }



}
